﻿using System;
using System.Linq;
using System.Threading;
using Cysharp.Threading.Tasks;
using Magify.Model;
using UnityEngine;
using UnityEngine.Networking;

namespace Magify.Tests
{
    internal class NetworkMoq : INetworkClient
    {
        internal delegate WebResponseMessage Handler(string url, WebRequestMessage message, CancellationToken cancellationToken);
        private readonly Handler _handler;

        public NetworkMoq(Handler handler)
        {
            _handler = handler;
        }

        public UniTask<WebResponseMessage> SendAsync(string url, WebRequestMessage message, CancellationToken cancellationToken)
        {
            try
            {
                var result = _handler(url, message, cancellationToken);
                return cancellationToken.IsCancellationRequested
                    ? UniTask.FromCanceled<WebResponseMessage>(cancellationToken)
                    : UniTask.FromResult(result);
            }
            catch (Exception e)
            {
                return UniTask.FromException<WebResponseMessage>(e);
            }
        }

        public static WebResponseMessage OkResult(WebRequestMessage message)
        {
            object text = string.Empty;
            if (message.Method == ServerApi.RequestTokenMethods.GetToken)
                text = new TokenResponse { Id = "id", Token = new string(Enumerable.Range(0, 100).Select(_ => 'a').ToArray()), };
            if (message.Method == ServerApi.MainNetworkRequestMethods.GetContext)
                text = new ContextResponse() { Context = new CampaignsContext() { ContextBuildTime = DateTime.UtcNow } };

            return new WebResponseMessage
            {
                Result = UnityWebRequest.Result.Success,
                ResponseCode = 200,
                Error = string.Empty,
                Text = JsonFacade.SerializeObject(text),
                RequestMessage = message,
            };
        }

        public static WebResponseMessage ProtocolErrorResult(WebRequestMessage message, string text)
        {
            return new WebResponseMessage
            {
                Result = UnityWebRequest.Result.ProtocolError,
                ResponseCode = 403,
                Error = string.Empty,
                Text = text,
                RequestMessage = message,
            };
        }

        public static WebResponseMessage ProtocolErrorResult(WebRequestMessage message, ErrorResponse text)
        {
            return new WebResponseMessage
            {
                Result = UnityWebRequest.Result.ProtocolError,
                ResponseCode = 403,
                Error = string.Empty,
                Text = JsonFacade.SerializeObject(text),
                RequestMessage = message,
            };
        }

        public static bool TryExtractEvent<T>(WebRequestMessage message, Magify.Model.EventType type, out T @event)
            where T: IAnalyticsEvent
        {
            if (message.Payload is StorePayload storePayload && storePayload.Model == type)
            {
                @event = (T)storePayload.Objects.FirstOrDefault();
                return @event != null;
            }
            @event = default;
            return false;
        }

        public static bool TryExtractToken(WebRequestMessage message, out AuthTokenPayload authTokenPayload)
        {
            authTokenPayload = message.Payload as AuthTokenPayload;
            return authTokenPayload != null;
        }

        public static bool TryExtractVerification(WebRequestMessage message, out VerifyPurchasePayload verifyPurchasePayload, RuntimePlatform platform = RuntimePlatform.Android)
        {
            if (message.Payload is VerifyPurchasePayload payload && message.Method == ServerApi.MainNetworkRequestMethods.VerifyPurchase(platform))
            {
                verifyPurchasePayload = payload;
                return true;
            }
            verifyPurchasePayload = default;
            return false;
        }
    }
}